import React, { LegacyRef, useEffect, useRef, useState } from 'react';
import { BarPlotTooltip } from '@graphpolaris/shared/lib/components/tooltip';
import {
  Axis,
  Bin,
  NumberValue,
  axisBottom,
  axisLeft,
  bin,
  extent,
  format,
  max,
  range,
  scaleBand,
  scaleLinear,
  select,
  Selection,
} from 'd3';

export type BarPlotProps = {
  data: { category: string; count: number }[];
  numBins: number;
  typeBarPlot: 'numerical' | 'categorical';
  marginPercentage?: {
    top: number;
    right: number;
    bottom: number;
    left: number;
  };
  className?: string;
  maxBarsCount?: number;
  strokeWidth?: number;
  axis?: boolean;
};

export const BarPlot = ({ typeBarPlot, numBins, data, marginPercentage, className, maxBarsCount, strokeWidth, axis }: BarPlotProps) => {
  const svgRef = useRef<SVGSVGElement | null>(null);
  const [dimensions, setDimensions] = useState({ width: 0, height: 0 });
  const [tooltipData, setTooltipData] = useState<{ x: number; y: number; content: { category: string; count: number } } | null>(null);

  useEffect(() => {
    if (!svgRef.current) return;

    const bounds = svgRef.current.getBoundingClientRect();
    const widthSVG: number = bounds.width;
    const heightSVG: number = bounds.height;

    setDimensions({ width: widthSVG, height: heightSVG });
    const svgPlot = select(svgRef.current);

    svgPlot.selectAll('*').remove();

    let groupMargin: Selection<SVGGElement, unknown, null, undefined>;
    let widthSVGwithinMargin: number = 0.0;
    let heightSVGwithinMargin: number = 0.0;

    // BarPlot for categorical data
    if (typeBarPlot == 'categorical') {
      if (!marginPercentage)
        marginPercentage = {
          top: 0.19,
          right: 0.02,
          bottom: 0.19,
          left: 0.19,
        };
      let margin = {
        top: marginPercentage.top * heightSVG,
        right: marginPercentage.right * widthSVG,
        bottom: marginPercentage.bottom * heightSVG,
        left: marginPercentage.left * widthSVG,
      };

      groupMargin = svgPlot.append('g').attr('transform', `translate(${margin.left},${margin.top})`);
      widthSVGwithinMargin = widthSVG - margin.left - margin.right;
      heightSVGwithinMargin = heightSVG - margin.top - margin.bottom;

      // Data processing
      const dataFiltered = data.filter((item) => item.category !== undefined);
      const dataSorted = dataFiltered.sort((a, b) => b.count - a.count);
      const dataTopCounts = dataSorted.filter((item, i) => !maxBarsCount || i < maxBarsCount);
      const maxCount = max(dataTopCounts, (d) => d.count) ?? 1;
      const xScale = scaleBand()
        .range([0, widthSVGwithinMargin])
        .domain(
          dataTopCounts.map(function (d) {
            return d.category;
          }),
        )
        .padding(0.2);

      // send this
      const yScale = scaleLinear().domain([0, maxCount]).range([heightSVGwithinMargin, 0]);

      // here out to axis component
      const yAxis1 = axisLeft(yScale).tickValues([0]).tickFormat(format('d')); // to show 0 without decimanls
      let yAxis2: Axis<NumberValue>;

      if (maxCount < 10) {
        yAxis2 = axisLeft(yScale).tickValues([maxCount]).tickFormat(format('d'));
      } else {
        yAxis2 = axisLeft(yScale).tickValues([maxCount]).tickFormat(format('.2s'));
      }

      groupMargin
        .selectAll<SVGRectElement, Bin<string, number>>('barplotCats')
        .data(dataTopCounts)
        .enter()
        .append('rect')
        .attr('x', (d) => xScale(d.category) || 0)
        .attr('y', (d) => yScale(d.count))
        .attr('width', xScale.bandwidth())
        .attr('height', (d) => heightSVGwithinMargin - yScale(d.count))
        .attr('class', 'hover:stroke-black hover:stroke-2 hover:stroke-secondary-400 fill-secondary-400')
        .on('mouseover', (event, d) => {
          setTooltipData({ x: event.layerX + 10, y: event.layerY - 28, content: d });
        })
        .on('mousemove', (event, d) => {
          if (tooltipData) {
            setTimeout(() => {
              setTooltipData((prevTooltipData) => {
                if (prevTooltipData) {
                  return {
                    ...prevTooltipData,
                    x: event.pageX + 10,
                    y: event.pageY - 28,
                  };
                }
                return prevTooltipData;
              });
            }, 16);
          }
        })
        .on('mouseleave', (d) => {
          setTooltipData(null);
        });

      if (axis) {
        groupMargin.append('g').call(yAxis1);
        groupMargin.append('g').call(yAxis2);
      }

      // Barplot for numerical data
    } else {
      if (!marginPercentage)
        marginPercentage = {
          top: 0.19,
          right: 0.02,
          bottom: 0.19,
          left: 0.19,
        };
      let margin = {
        top: marginPercentage.top * heightSVG,
        right: marginPercentage.right * widthSVG,
        bottom: marginPercentage.bottom * heightSVG,
        left: marginPercentage.left * widthSVG,
      };

      groupMargin = svgPlot.append('g').attr('transform', `translate(${margin.left},${margin.top})`);
      widthSVGwithinMargin = widthSVG - margin.left - margin.right;
      heightSVGwithinMargin = heightSVG - margin.top - margin.bottom;
      const dataCount = data.map((obj) => obj.count);

      const extentData = extent(dataCount);
      const [min, max] = extentData as [number, number];

      const xScale = scaleLinear().range([0, widthSVGwithinMargin]).domain([min, max]); //.nice();

      // histogram -> deprecated
      // On bin(), .thresholds(x.ticks(numBins)). Creates artifacts: not full rects on the plot
      // CHECK THIS: https://dev.to/kevinlien/d3-histograms-and-fixing-the-bin-problem-4ac5

      const rangeTH: number[] = range(min, max, (max - min) / numBins);

      if (rangeTH.length != numBins) {
        rangeTH.pop();
      }

      const histogram = bin()
        .value(function (d) {
          return d;
        })
        .domain([min, max])
        .thresholds(rangeTH);

      const bins = histogram(dataCount);
      const extentBins: [number, number] = extent(bins, (d) => d.length) as [number, number];

      const yScale = scaleLinear()
        .range([heightSVGwithinMargin, 0])
        .domain([0, extentBins[1]] as [number, number]);

      groupMargin
        .selectAll<SVGRectElement, Bin<number, number>>('barplotbins')
        .data(bins)
        .enter()
        .append('rect')
        .attr('x', 0)
        .attr('stroke', 'none')
        .attr('fill', 'hsl(var(--clr-sec--400))')
        .attr('transform', (d) => 'translate(' + xScale(d.x0 || 0) + ',' + yScale(d.length) + ')')
        .attr('width', (d) => xScale(d.x1 || 0) - xScale(d.x0 || 0))
        .attr('height', (d) => heightSVGwithinMargin - yScale(d.length));

      const xAxis = axisBottom(xScale)
        //.tickValues([Math.round((min + max) / 2.0), max])
        .tickValues([max])
        .tickFormat(format('.2s'));

      let xAxis2: Axis<NumberValue>;

      if (min < 10) {
        xAxis2 = axisBottom(xScale).tickValues([min]).tickFormat(format('d'));
      } else {
        xAxis2 = axisBottom(xScale).tickValues([min]).tickFormat(format('.2s'));
      }

      groupMargin
        .append('g')
        .attr('transform', 'translate(0,' + heightSVGwithinMargin + ')')
        .call(xAxis2);

      groupMargin
        .append('g')
        .attr('transform', 'translate(0,' + heightSVGwithinMargin + ')')
        .call(xAxis);

      const yAxis1 = axisLeft(yScale).tickValues([0]).tickFormat(format('d')); // to show 0 without decimanls
      //const yAxis2 = axisLeft(yScale).tickValues([extentBins[1]]).tickFormat(format('.2s'));

      let yAxis2: Axis<NumberValue>;

      if (extentBins[1] < 10) {
        yAxis2 = axisLeft(yScale).tickValues([extentBins[1]]).tickFormat(format('d'));
      } else {
        yAxis2 = axisLeft(yScale).tickValues([extentBins[1]]).tickFormat(format('.2s'));
      }

      if (axis) {
        // two axis for each number, it solves the 0.0 problem with ".2s" notatation
        groupMargin.append('g').call(yAxis1);
        groupMargin.append('g').call(yAxis2);
      }
    }

    svgPlot.selectAll('.domain').style('stroke', 'hsl(var(--clr-sec--400))');
    svgPlot.selectAll('.tick line').style('stroke', 'hsl(var(--clr-sec--400))');
    svgPlot.selectAll('.tick text').attr('class', 'font-mono').style('stroke', 'none').style('fill', 'hsl(var(--clr-sec--500))');

    groupMargin
      .append('rect')
      .attr('x', 0.0)
      .attr('y', 0.0)
      .attr('width', widthSVGwithinMargin)
      .attr('height', heightSVGwithinMargin)
      .attr('rx', 0)
      .attr('ry', 0)
      .attr('stroke-width', strokeWidth || 0)
      .attr('fill', 'none')
      .attr('stroke', 'hsl(var(--clr-sec--400))');
  }, [data, svgRef]);

  return (
    <div className={!!className ? className : ''}>
      <svg ref={svgRef} className="" width="100%" height="100%"></svg>
      {tooltipData && (
        <BarPlotTooltip x={tooltipData.x} y={tooltipData.y}>
          <div>
            <span>{tooltipData.content.category.toString()}</span>: <span>{tooltipData.content.count}</span>
          </div>
        </BarPlotTooltip>
      )}
    </div>
  );
};

export default BarPlot;
