import * as React from 'react';
import {
  useFloating,
  autoUpdate,
  offset,
  flip,
  shift,
  hide,
  arrow,
  useHover,
  useFocus,
  useDismiss,
  useRole,
  useInteractions,
  useMergeRefs,
  FloatingPortal,
  FloatingArrow
} from '@floating-ui/react';
import type { Placement } from '@floating-ui/react';
import { FloatingDelayGroup } from '@floating-ui/react';

interface TooltipOptions {
  initialOpen?: boolean;
  placement?: Placement;
  open?: boolean;
  onOpenChange?: (open: boolean) => void;
  boundaryElement?: React.RefObject<HTMLElement> | null;
  showArrow?: boolean;
}

export function useTooltip({
  initialOpen = false,
  placement = 'top',
  open: controlledOpen,
  onOpenChange: setControlledOpen,
  boundaryElement = null,
  showArrow = false
}: TooltipOptions = {}): {
  open: boolean;
  setOpen: (open: boolean) => void;
  interactions: ReturnType<typeof useInteractions>;
  data: ReturnType<typeof useFloating>;
} {
  const [uncontrolledOpen, setUncontrolledOpen] = React.useState(initialOpen);

  const open = controlledOpen ?? uncontrolledOpen;
  const setOpen = setControlledOpen ?? setUncontrolledOpen;
  const arrowRef = React.useRef<SVGSVGElement | null>(null);

  let config = {
    placement,
    open,
    onOpenChange: setOpen,
    whileElementsMounted: autoUpdate,
    middleware: [
      offset(5),
      flip({
        crossAxis: placement.includes('-'),
        fallbackAxisSideDirection: 'start',
        padding: 5
      }),
      shift({ padding: 5 }),
    ],
  }

  if (boundaryElement != null) {
    const boundary = boundaryElement?.current ?? undefined;
    config.middleware.find(x => x.name == 'flip')!.options[0].boundary = boundary;
    config.middleware.find(x => x.name == 'shift')!.options[0].boundary = boundary;
    config.middleware.push(hide({ boundary }));
  }

  if (showArrow) {
    config.middleware.push(arrow({ element: arrowRef }));
  }

  const data = useFloating(config);
  (data.refs as any).arrow = arrowRef;

  const context = data.context;

  const hover = useHover(context, {
    move: false,
    enabled: controlledOpen == null,
  });
  const focus = useFocus(context, {
    enabled: controlledOpen == null,
  });
  const dismiss = useDismiss(context);
  const role = useRole(context, { role: 'tooltip' });

  const interactions = useInteractions([hover, focus, dismiss, role]);

  return React.useMemo(
    () => ({
      open,
      setOpen,
      interactions: interactions,
      data: data,
    }),
    [open, setOpen, interactions, data],
  );
}

type ContextType = ReturnType<typeof useTooltip> | null;

const TooltipContext = React.createContext<ContextType>(null);

export const useTooltipContext = () => {
  const context = React.useContext(TooltipContext);

  if (context == null) {
    throw new Error('Tooltip components must be wrapped in <Tooltip />');
  }

  return context;
};

export function Tooltip({ children, ...options }: { children: React.ReactNode } & TooltipOptions) {
  // This can accept any props as options, e.g. `placement`,
  // or other positioning options.
  const tooltip = useTooltip(options);
  
  return <TooltipContext.Provider value={tooltip}>
    {children}
  </TooltipContext.Provider>;
}

export const TooltipTrigger = React.forwardRef<HTMLElement, React.HTMLProps<HTMLElement> & { asChild?: boolean, x?: number, y?: number }>(function TooltipTrigger(
  { children, asChild = false, x = null, y = null, ...props },
  propRef,
) {
  const context = useTooltipContext();
  const childrenRef = React.useMemo(() => {
    if (children == null) {
      return null;
    } else {
      return (children as any).ref;
    }
  }, [children]);

  const ref = useMergeRefs([context.data.refs.setReference, propRef, childrenRef]);

  React.useEffect(() => {
    if (x && y && context.data.refs.reference.current != null) {
      const element = context.data.refs.reference.current as HTMLElement;
      element.style.position = 'absolute';
      const {x: offsetX, y: offsetY} = element.getBoundingClientRect();
      element.getBoundingClientRect = () => {
        return {
          width: 0,
          height: 0,
          x: offsetX,
          y: offsetY,
          top: y + offsetY,
          left: x + offsetX,
          right: x + offsetX,
          bottom: y + offsetY,
        } as DOMRect
      }
      context.data.update();
    }
  }, [x, y]);

  // `asChild` allows the user to pass any element as the anchor
  if (asChild && React.isValidElement(children)) {
    return React.cloneElement(
      children,
      context.interactions.getReferenceProps({
        ref,
        ...props,
        ...children.props,
        'data-state': context.open ? 'open' : 'closed',
      }),
    );
  }

  return (
    <div
      ref={ref}
      // The user can style the trigger based on the state
      data-state={context.open ? 'open' : 'closed'}
      {...context.interactions.getReferenceProps(props)}
    >
      {children}
    </div>
  );
});

export const TooltipContent = React.forwardRef<
  HTMLDivElement,
  React.HTMLProps<HTMLDivElement> & {
    side?: string;
  }
>(function TooltipContent({ style, className, ...props }, propRef) {
  const context = useTooltipContext();
  const ref = useMergeRefs([context.data.refs.setFloating, propRef]);

  if (!context.open) return null;

  return (
    <FloatingPortal>
      <div
        ref={ref}
        className={`z-50 max-w-64 rounded bg-light px-2 py-1 shadow text-xs border border-secondary-200 text-dark animate-in fade-in-0 zoom-in-95 data-[state=closed]:animate-out data-[state=closed]:fade-out-0 data-[state=closed]:zoom-out-95 data-[side=bottom]:slide-in-from-top-2 data-[side=left]:slide-in-from-right-2 data-[side=right]:slide-in-from-left-2 data-[side=top]:slide-in-from-bottom-2${className ? ` ${className}` : ''}`}
        style={{
          ...context.data.floatingStyles,
          ...style,
          display: context.data.middlewareData.hide?.referenceHidden ? 'none' : 'block',
        }}
        {...context.interactions.getFloatingProps(props)}
      >
        { props.children }
        { context.data.middlewareData.arrow ? <FloatingArrow 
          ref={(context.data.refs as any).arrow} 
          context={context.data.context} 
          style={{fill: 'white'}}
        /> : null }
      </div>
    </FloatingPortal>
  );
});

export const TooltipProvider = (props: { delayDuration: number; children: React.ReactNode }) => {
  return <FloatingDelayGroup delay={props.delayDuration}>{props.children}</FloatingDelayGroup>;
};
